package hu.elte.inf.teralg.graph;

import java.io.FileReader;
import java.io.IOException;
import java.util.*;

/**
 * Térinformatikai algoritmusok
 * 5. beadandó: Legolcsóbb út
 * @author Máté Cserép <mcserep@inf.elte.hu>
 */
public class Dijkstra {
    public static void main(String[] args) {
        int n = 0, e = 0;           // csúcsok és élek száma
        String[] cities = null;     // városok (csúcsok) neve
        Graph graph = new Graph();  // gráf

        // Adatok beolvasása
        try {
            FileReader fr = new FileReader("flights2.txt");
            Scanner sc = new Scanner(fr);

            // Csúcsok beolvasása
            n = sc.nextInt();
            cities = new String[n];
            for(int i = 0; i < n; ++i) {
                cities[i] = sc.next();
                graph.add(cities[i]);
            }
            Arrays.sort(cities);

            // Élek beolvasása
            e = sc.nextInt();
            for(int i = 0; i < e; ++i) {
                String cityA = sc.next();
                String cityB = sc.next();
                Integer cost = sc.nextInt();

                graph.set(cityA, cityB, cost);
                graph.set(cityB, cityA, cost);
            }
            sc.close();
            fr.close();
        }
        catch(IOException ex) {
            System.out.println("File handling error: " + ex.getMessage());
        }

        if(cities.length == 0) {
            System.out.println("Üres bemenet.");
        }
        else if(Arrays.binarySearch(cities, "Budapest") < 0) {
            System.out.println("Budapest nincs a városok között.");
        }
        else {
            PriorityQueue<Item> process = new PriorityQueue<Item>();  // kiterjesztendő csúcsok sora
            Set<String> ready = new HashSet<String>();
            Map<String, Integer> distance = new HashMap<String, Integer>();
            Map<String, String> parent = new HashMap<String, String>();

            // Start csúcs inicializálása
            process.add(new Item("Budapest", 0));
            distance.put("Budapest", 0);
            parent.put("Budapest", null);

            // Iteratív kiterjesztés
            while(!process.isEmpty()) {
                // Aktuálisan kiterjesztendő csúcs
                Item current = process.poll();
                String source = current.name;
                ready.add(source);

                // Szomszédok frissítése
                for(String target : graph.neighbours(source)) {
                    if(!ready.contains(target)) {
                        // Eddigi költség
                        Integer oldCost = distance.get(target);
                        // Új költség
                        Integer newCost = distance.get(source) + graph.get(source, target);

                        if (!distance.containsKey(target)) {
                            distance.put(target, newCost);
                            parent.put(target, source);
                            process.add(new Item(target, newCost));
                        } else if (newCost < distance.get(target)) {
                            process.remove(new Item(target, oldCost));
                            distance.put(target, newCost);
                            parent.put(target, source);
                            process.add(new Item(target, newCost));
                        }
                    }
                }
            }

            // Eredmény kiírása
            for(String city : cities) {
                System.out.println(city + ": " + distance.get(city) + " Ft");

                System.out.print(city);
                city = parent.get(city);
                while(city != null) {
                    System.out.print(" <- " + city);
                    city = parent.get(city);
                }
                System.out.println();
            }
        }
    }

    private static class Item implements Comparable<Item> {
        public String name;
        public Integer distance;

        public Item(String name, Integer distance) {
            this.name = name;
            this.distance = distance;
        }

        @Override
        public int compareTo(Item o) {
            return distance.compareTo(o.distance);
        }
    }
}
